#include <windows.h>
#include <stdio.h>

#include "dxfile.h"
#include "rmxftmpl.h"
#include "rmxfguid.h"

#define ReleaseCOM(x) if(x) { x->Release(); x = NULL; }

typedef struct sFrame
{
  char *Name;       // Name of frame
  sFrame *Child;    // Child linked list
  sFrame *Sibling;  // Sibling linked list
  
  sFrame()          // constructor
  {
    Name  = NULL;
    Child = Sibling = NULL;
  }

  ~sFrame()         // destructor
  {
    delete Name;
    delete Child;
    delete Sibling;
    Name   = NULL;
    Child  = Sibling = NULL;
  }
} sFrame;


int PASCAL WinMain(HINSTANCE hInst, HINSTANCE hPrev, LPSTR szCmdLine, int nCmdShow);

BOOL ParseXFile(char *Filename);
void ParseXFileData(IDirectXFileData *pData, sFrame *ParentFrame);

int PASCAL WinMain(HINSTANCE hInst, HINSTANCE hPrev, LPSTR szCmdLine, int nCmdShow)
{
  ParseXFile("sample.x");
  return FALSE;
}

BOOL ParseXFile(char *Filename)
{
  IDirectXFile           *pDXFile = NULL;
  IDirectXFileEnumObject *pDXEnum = NULL;
  IDirectXFileData       *pDXData = NULL;

  // Create the .X file object
  if(FAILED(DirectXFileCreate(&pDXFile)))
    return FALSE;

  // Register the templates in use
  // Use the standard retained mode templates from Direct3D
  if(FAILED(pDXFile->RegisterTemplates((LPVOID)D3DRM_XTEMPLATES, D3DRM_XTEMPLATE_BYTES))) {
    pDXFile->Release();
    return FALSE;
  }

  // Create an enumeration object
  if(FAILED(pDXFile->CreateEnumObject((LPVOID)Filename, DXFILELOAD_FROMFILE, &pDXEnum))) {
    pDXFile->Release();
    return FALSE;
  }

  // Create a root frame
  sFrame *ParentFrame = new sFrame();
  ParentFrame->Name = new char[7];
  strcpy(ParentFrame->Name, "$ROOT$");

  // Enumerate all top-level templates
  while(SUCCEEDED(pDXEnum->GetNextDataObject(&pDXData))) {
    ParseXFileData(pDXData, ParentFrame);
    ReleaseCOM(pDXData);
  }

  // Release objects
  ReleaseCOM(pDXEnum);
  ReleaseCOM(pDXFile);

  // Delete the root when no longer needed
  delete ParentFrame;

  // Return a success
  return TRUE;
}

void ParseXFileData(IDirectXFileData *pData, sFrame *ParentFrame)
{
  IDirectXFileObject *pSubObj  = NULL;
  IDirectXFileData   *pSubData = NULL;
  IDirectXFileDataReference *pDataRef = NULL;
  const GUID *pType = NULL;
  char       *pName = NULL;
  DWORD       dwSize;
  sFrame     *Frame = NULL;
  sFrame     *SubFrame = NULL;
  
  // Get the template type
  if(FAILED(pData->GetType(&pType)))
    return;

  // Get the template name (if any)
  if(FAILED(pData->GetName(NULL, &dwSize)))
    return;
  if(dwSize) {
    if((pName = new char[dwSize]) != NULL)
      pData->GetName(pName, &dwSize);
  }

  // Give template a default name if none found
  if(pName == NULL) {
    if((pName = new char[9]) == NULL)
      return;
    strcpy(pName, "Template");
  }

  // Set sub frame parent
  SubFrame = ParentFrame;

  // Process the frame templates
  if(*pType == TID_D3DRMFrame) {
    // Create a frame
    Frame = new sFrame();

    // Store the name
    Frame->Name = pName;
    pName = NULL;

    // Add to parent frame as sibling
    Frame->Sibling = ParentFrame->Child;
    ParentFrame->Child = Frame;

    // Set sub frame parent
    SubFrame = Frame;

    // Display a message describing which frame it was added to
    char Buffer[1024];
    sprintf(Buffer, "Child frame: %s\r\nParent frame: %s", Frame->Name, ParentFrame->Name);
    MessageBox(NULL, Buffer, "Added Frame", MB_OK);
  }

  // Scan for embedded templates
  while(SUCCEEDED(pData->GetNextObject(&pSubObj))) {

    // Process embedded references
    if(SUCCEEDED(pSubObj->QueryInterface(IID_IDirectXFileDataReference, (void**)&pDataRef))) {
      if(SUCCEEDED(pDataRef->Resolve(&pSubData))) {
        ParseXFileData(pSubData, SubFrame);
        ReleaseCOM(pSubData);
      }
      ReleaseCOM(pDataRef);
    }

    // Process non-referenced embedded templates
    if(SUCCEEDED(pSubObj->QueryInterface(IID_IDirectXFileData, (void**)&pSubData))) {
      ParseXFileData(pSubData, SubFrame);
      ReleaseCOM(pSubData);
    }
    ReleaseCOM(pSubObj);
  }

  // Release name buffer
  delete pName;
}
